// AnalogOutputs.js
//  This module handles analog output processing and eventing

//#region Dependencies, Constants and Globals
//#region Dependencies
var fs = require('fs');
var moment = require('moment');
var exec = require('child_process').exec;
//#endregion Dependencies

//#region Constants
// Define the configuraton for the I/O
var ANALOG_OUTPUTS_LOCATION = '/sys/class/hwmon/hwmon2';
var ANALOG_OUTPUTS_POINTS = 4;//default = 4
var ANALOG_OUTPUTS_DEFAULT_RANGE = 20;// default = 20
var ANALOG_OUTPUT_CONVERT_TO_RAW_SLOPE = 200;
var ANALOG_OUTPUT_CONVERT_TO_RAW_OFFSET = 0;

// SNMP - this is here for now, could move
var AuthProtocol = ["None", "MD5", "SHA"];
var PrivProtocol = ["None", "DES", "AES"];
//#endregion Constants

//#region Globals
// The following array holds, in memory, the current value (in raw bit values - 0 to 2048)
var analogOutputsArray = [
    { "Address" : "0", "Name" : "Name",  "RawValue": 0, "RangedValue": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": 20, "TimeChanged": 0 }, 
    { "Address" : "1", "Name" : "Name",  "RawValue": 0, "RangedValue": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": 20, "TimeChanged": 0 }, 
    { "Address" : "2", "Name" : "Name",  "RawValue": 0, "RangedValue": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": 20, "TimeChanged": 0 }, 
    { "Address" : "3", "Name" : "Name",  "RawValue": 0, "RangedValue": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": 20, "TimeChanged": 0 }
];

var SummitInfo = {};
var SmtpManager = {};
var SnmpManager = {};
var SqliteManager = {};
var LogManager = {};
var SocketIoManager = {};
//#endregion Globals
//#endregion Dependencies, Constants and Globals

//#region Analog Outputs Initialization
// Initialize the analog outputs upon system start
//      - Checks if database table exists, if not creates the table and fills it with default data for each point
//      - Else reads from the database and updates analogOutputsArray
exports.InitializeAnalogOutputs = function(info, smtp, snmp, db, io, logger, callback) {
    // Initialize globals
    SummitInfo = info;
    SmtpManager = smtp;
    SnmpManager = snmp;
    SqliteManager = db;
    SocketIoManager = io;
    LogManager = logger;
    
    // Try getting the AnalogOutputs table
    SqliteManager.get("SELECT name FROM sqlite_master WHERE type='table' AND name='AnalogOutputs'", function (err, rows) {
        if (err !== null) {
            LogManager.error('Analog Outputs - Initialize Error (Check)=' + JSON.stringify(err));
        }
        else if (typeof rows === "undefined") {
            // We read nothing, so let's create the table ...
            LogManager.info("Analog Outputs - Initialize by creating database table ...");
            SqliteManager.exec("CREATE TABLE if not exists AnalogOutputs (Address INTEGER, Name TEXT, RawValue INTEGER, RangedValue TEXT, Enabled TEXT, SendEvents TEXT, Range TEXT, TimeChanged TEXT )", function (err) {
                if (err !== null) {
                    LogManager.error('Analog Outputs - Initialize Error (Create)=' + JSON.stringify(err));
                }
                else {
                    var timeChanged = moment();// Record the create time as the last time the I/O point value was changed
                    
                    // Created the table, now add data ...
                    SqliteManager.run("BEGIN")
                    for (var i = 0; i < ANALOG_OUTPUTS_POINTS; i++) {
                        // Initialize the array
                        UpdateAnalogOutputConfiguration(i, "Analog Ouput " + (i + 1), 0, 0, "FALSE", "FALSE", ANALOG_OUTPUTS_DEFAULT_RANGE, timeChanged);
                        // Empty space here to be consistent
                        // Empty space here to be consistent
                        
                        // Initialize the database
                        insertStatement = i + ", 'Analog Output " + (i + 1) + "', 0, '0', 'FALSE', 'FALSE', '" + ANALOG_OUTPUTS_DEFAULT_RANGE + "', '" + timeChanged + "'";
                        SqliteManager.run("INSERT INTO AnalogOutputs (Address, Name, RawValue, RangedValue, Enabled, SendEvents, Range, TimeChanged) VALUES (" + insertStatement + ")", function (err) {
                            if (err !== null) {
                                LogManager.error('Analog Outputs - Initialize Error (Insert)=' + JSON.stringify(err));
                            }
                        });
                    }
                    SqliteManager.run("COMMIT");

                    // Analog Outputs table created
                    if (typeof callback == "function") {
                        LogManager.info('Analog Outputs - Initialized new database table');
                        callback();
                    }
                }
            });
        }
        else {
            LogManager.info("Analog Outputs - Initialize by reading database ...");
            // Write out current AnalogOutputs table
            SqliteManager.each("SELECT rowid as id, Address, Name, RawValue, RangedValue, Enabled, SendEvents, Range, TimeChanged FROM AnalogOutputs ORDER BY Address", function (err, row) {
                if (err !== null) {
                    LogManager.error('Analog Outputs - Initialize Error (Select)=' + JSON.stringify(err));
                }
                else {
                    UpdateAnalogOutputConfiguration(row.Address, row.Name, row.RawValue, row.RangedValue, row.Enabled, row.SendEvents, row.Range, row.TimeChanged)
                }
            }, function () { 
                // Analog Outputs read from table
                if (typeof callback == "function") {
                    LogManager.info('Analog Outputs - Initialized from existing database table');
                    callback();
                }
            });
        }
    });
};

// Added this in later to support Apex4 Integration
var Apex4Token = null;
var Apex4Manager = null;
exports.ConfigureApex4TokenAndSocket = function (apex4Token, apex4Socket) {
    Apex4Token = apex4Token;
    Apex4Manager = apex4Socket;
}

exports.InitializeVoltageFormula = function(voltageFormulae,logger) {
    logger.info("analogoutputs.InitializeVoltageFormulae: Setting voltage formulae=" + JSON.stringify(voltageFormulae));
    ANALOG_OUTPUT_CONVERT_TO_RAW_SLOPE = Number(voltageFormulae.slope);
    ANALOG_OUTPUT_CONVERT_TO_RAW_OFFSET = Number(voltageFormulae.offset);
}
//#endregion Analog Outputs Initialization

//#region Analog Outputs API
// Adjust the analogOutputsArray
module.exports.UpdateAnalogOutputConfiguration = UpdateAnalogOutputConfiguration;
function UpdateAnalogOutputConfiguration(Address, Name, RawValue, RangedValue, Enabled, SendEvents, Range, TimeChanged) {
    analogOutputsArray[Address].Name = Name;
    analogOutputsArray[Address].RawValue = RawValue;
    analogOutputsArray[Address].RangedValue = RangedValue;
    analogOutputsArray[Address].Enabled = Enabled;
    analogOutputsArray[Address].SendEvents = SendEvents;
    analogOutputsArray[Address].Range = Range;
    analogOutputsArray[Address].TimeChanged = TimeChanged;
    sendValueAnalogOutputs();// Update the Analog Outputs configuration
}

exports.ResetAnalogOutputs = function(ledManager, callback) {
    var timeChanged = moment();
    for (index = 0; index < ANALOG_OUTPUTS_POINTS; index++) {
        // Reset the database point ...
        // We may not have to do this database update
        //db.run("UPDATE AnalogOutputs SET RawValue='" + 0 + "', TimeChanged='" + timeChanged + "' WHERE Address = '" + index + "'", 
        //    sendSocketSnmpSmtp(analogOutputsArray[index], db, io, logger, netSnmp));
        
        // Reset the output point ...        
        fs.writeFile(ANALOG_OUTPUTS_LOCATION + "/out" + index + "_output", '0', function (err) {
            if (err) {
                return LogManager.error('Analog Outputs - Reset Error (Write)=' + JSON.stringify(err));
            }
        });
    }

    // Analog Outputs reset is complete
    if (typeof callback == "function") {
        LogManager.debug('Analog Outputs - Reset to initial values');
        callback();
    }
}


// Get DigitalInputsArray
module.exports.GetAnalogOutputsArray = GetAnalogOutputsArray;
function GetAnalogOutputsArray() {
    return analogOutputsArray;
}
//#endregion Analog Outputs API

//#region Analog Outputs Processing
var processCounter = 0;
var processTotalTime = 0;
exports.ProcessAnalogOutputs = function (ledManager) {
    //timeStart = moment();
    
    var index = 0;// This is an index to analog I/O point 
    var fileIndex = 0;// This is an index to the GPIO location
    var timeChanged = moment();
    for (index = 0; index < ANALOG_OUTPUTS_POINTS; index++) {
        fileIndex = index;
        
        // Only read value if the I/O point is enabled
        if (analogOutputsArray[index].Enabled === "TRUE") {
            try {
                data = fs.readFileSync(ANALOG_OUTPUTS_LOCATION + "/out" + fileIndex + "_output");
                readValue = String(data).replace(/(\r\n|\n|\r)/gm, "");
                
                // Check to see if value has changed, if so update database and clients ...
                if (analogOutputsArray[index].RawValue != readValue) {
                    analogOutputsArray[index].RawValue = readValue;
                    analogOutputsArray[index].TimeChanged = timeChanged;
                    sendSocketSnmpSmtp(analogOutputsArray[index]);

                    // Value has changed, update the database and then the clients ...
                    SqliteManager.run("UPDATE AnalogOutputs SET RawValue='" + readValue 
                        + "', TimeChanged='" + timeChanged 
                        + "' WHERE Address = '" + index + "'", function () {
                    });
                }
                //logger.info("Analog Output[" + (index+1) + "]=" + readValue + " (Enabled)")
            }
			catch (error) {
                LogManager.error('Analog Outputs - Process Error (Read)=' + JSON.stringify(error));
            }
        }
        else {
            //logger.info("Analog Output[" + (index+1) + "]=" + analogOutputsArray[index].Value + " (Disabled)")
        }
    }
    //if (processCounter === 500) {
    //    logger.info("ProcessAnalogOutputs: Average process time = " + (processTotalTime / processCounter).toFixed(2) + "msec");
    //    processCounter = 0;
    //    processTotalTime = 0;
    //}
    //processCounter++;
    //processTotalTime += (Date.now() - timeStart);
}
//#endregion Analog Outputs Processing

//#region Analog Outputs Eventing
function sendSocketSnmpSmtp(analogOutput) {
    rangedValue = ((analogOutputsArray[analogOutput.Address].RawValue - ANALOG_OUTPUT_CONVERT_TO_RAW_OFFSET)/ ANALOG_OUTPUT_CONVERT_TO_RAW_SLOPE).toFixed(2)
    LogManager.info("Analog Outputs - [Value UPDATED] Analog Output[" + (parseInt(analogOutput.Address) + 1) + "] '" + analogOutput.Name + "' Value=" + rangedValue);
    LogManager.debug("Analog Outputs - [Value UPDATED] Analog Output[" + (parseInt(analogOutput.Address) + 1) + "] =" + JSON.stringify(analogOutput));

    // SOCKET.IO: Send changed value to all attached clients
    sendValueAnalogOutput("AnalogOutput", analogOutput);
    
    // SNMP: Send SNMP Trap/Notification if threshold(s) crossed
    sendSNMP(analogOutput);

    // SMTP: Send SMTP message if threshold(s) crossed
    sendSMTP(analogOutput)

    // APEX4: Send message to Apex 4
    if (Apex4Manager) {
        var changestate = {};
        var uid = "AnalogOutput" + analogOutput.Address;
        changestate["token"] = analogOutput;
        changestate["time"] = analogInput.TimeChanged;
        changestate[uid] = analogOutput.RangedValue;
        Apex4Manager.emit("changestate", changestate);
        LogManager.info("APEX4 message=" + JSON.stringify(changestate));
    }
}

function sendValueAnalogOutput(dataType, data) {
    SocketIoManager.sockets.emit(dataType, data);
}

function sendValueAnalogOutputs() {
    enabled = 0;
    for (i = 0; i < analogOutputsArray.length; i++) {
        if (analogOutputsArray[i].Enabled === "TRUE") enabled++;
    }
    SocketIoManager.sockets.emit("AnalogOutputs", { "AnalogOutputsTotal": analogOutputsArray.length, "AnalogOutputsEnabled": enabled })
}

function sendSNMP(analogOutput) {
    // SNMP definitions - also refer to the TASC-SUMMIT.MIB file for detailed information
    if (analogOutput.Enabled == "FALSE") return;
    if (analogOutput.SendEvents == "FALSE") return;
    
    address = parseInt(analogOutput.Address) + 1;// To overcome concatenation
    LogManager.trace("Analog Outputs- Sending SNMP Trap for Analog Output[" + address + "] '" + analogOutput.Name + "'");


    // Define the notification
    var snmpCommand;
    var snmpHost;
    var snmpNotification = "";
    snmpNotification += "1.3.6.1.4.1.11902.3.4.1000.1000.0." + (address - 1 + 1000) + " ";
    snmpNotification += "1.3.6.1.4.1.11902.3.4." + (address) + ".0.1 s '" + moment().toString() + "' ";
    snmpNotification += "1.3.6.1.4.1.11902.3.4." + (address) + ".0.2 i " + (parseInt(analogOutput.Address) + 1) + " ";
    snmpNotification += "1.3.6.1.4.1.11902.3.4." + (address) + ".0.3 s '" + analogOutput.Name + "' ";
    snmpNotification += "1.3.6.1.4.1.11902.3.4." + (address) + ".0.4 s '" + (((analogOutput.RawValue - ANALOG_OUTPUT_CONVERT_TO_RAW_OFFSET) / ANALOG_OUTPUT_CONVERT_TO_RAW_SLOPE).toFixed(2)).toString() + "' ";
    snmpNotification += "1.3.6.1.4.1.11902.3.4." + (address) + ".0.20 s '" + SummitInfo.MACID + "' ";
    snmpNotification += "1.3.6.1.4.1.11902.3.4." + (address) + ".0.21 s '" + SummitInfo.Name + "' ";

    // Send the SNMP notification
    for (i = 0; i < SnmpManager.length; i++) {
        // Skip if this host is not enabled
        if (SnmpManager[i].Enabled === "FALSE") continue;

        // Set up the SNMP trap command
        snmpCommand = "snmptrap ";
        if (SnmpManager[i].SnmpVersion == 2) {
            snmpCommand += "-v 2c ";
            snmpCommand += "-c " + SnmpManager[i].Community + " ";
        }
        else if (SnmpManager[i].SnmpVersion == 3) {
            snmpCommand += "-v 3 ";
            var EngineId = "0x" + SnmpManager[i].Community.replace(/\-/g, '');
            snmpCommand += "-e " + EngineId + " ";
            snmpCommand += "-u " + SnmpManager[i].AuthUser + " ";
            if (SnmpManager[i].AuthProtocol > 0) snmpCommand += "-a " + AuthProtocol[SnmpManager[i].AuthProtocol] + " -A " + SnmpManager[i].AuthPassword + " ";
            if (SnmpManager[i].PrivProtocol > 0) snmpCommand += "-x " + PrivProtocol[SnmpManager[i].PrivProtocol] + " -X " + SnmpManager[i].PrivPassword + " ";
            var securityLevel = "-l noAuthNoPriv ";
            if (SnmpManager[i].AuthProtocol > 0 && SnmpManager[i].PrivProtocol == 0)
                securityLevel = "-l authNoPriv ";
            if (SnmpManager[i].AuthProtocol > 0 && SnmpManager[i].PrivProtocol > 0)
                securityLevel = "-l authPriv ";
            snmpCommand += securityLevel;
        } else continue;
        
        // Send trap to host ...
        snmpHost = SnmpManager[i].IpAddress + " '' ";
        LogManager.info("Analog Outputs - Sending SNMP Event to '" + SnmpManager[i].HostName + "' for Analog Output[" + address + "] '" + analogOutput.Name + "' Value=" + (((analogOutput.RawValue - ANALOG_OUTPUT_CONVERT_TO_RAW_OFFSET) / ANALOG_OUTPUT_CONVERT_TO_RAW_SLOPE).toFixed(2)).toString());
        LogManager.trace("Analog Outputs - SendSNMP: " + snmpCommand + snmpHost + snmpNotification);
        exec(snmpCommand + snmpHost + snmpNotification, function (error, stdout, stderr) {
            if (error !== null) {
                LogManager.error("Analog Outputs - SendSNMP. " + error);
            } else {
                LogManager.debug("Analog Outputs - SendSNMP. " + stdout);
            }
        });
    }
}

function sendSMTP(analogOutput) {
    // SNMP definitions - also refer to the TASC-SUMMIT.MIB file for detailed information
    if (analogOutput.Enabled == "FALSE") return;
    if (analogOutput.SendEvents == "FALSE") return;
    // NOTE: Decision to send email is decided within SMTP module
    
    // Set up the SMTP Email message
    var message = [];
    address = parseInt(analogOutput.Address) + 1;// To overcome concatenation
    var event = "'" + analogOutput.Name + "' has changed to value=" + (((analogOutput.RawValue - ANALOG_OUTPUT_CONVERT_TO_RAW_OFFSET) / ANALOG_OUTPUT_CONVERT_TO_RAW_SLOPE).toFixed(2)).toString()+"V";
    message += moment().toString() + "\n";
    message += "Summit MACID=" + SummitInfo.MACID + "\n";
    message += "Summit Name=" + SummitInfo.Name + "\n";
    message += "Address=" + (parseInt(analogOutput.Address) + 1) + "\n";
    message += "Name=" + analogOutput.Name + "\n";
    message += "Value=" + (((analogOutput.RawValue - ANALOG_OUTPUT_CONVERT_TO_RAW_OFFSET) / ANALOG_OUTPUT_CONVERT_TO_RAW_SLOPE).toFixed(2)).toString() + "\n";
    
    var ioInfo = "Analog Output[" + address + "] '" + analogOutput.Name + "'";
    
    LogManager.info("Analog Outputs - Sending SMTP Event for Analog Output[" + address + "] '" + analogOutput.Name + "'");
    SmtpManager.SendSmtp(message, ioInfo, event, function (result) {
        LogManager.trace("Analog Outputs - SMTP Email Event Response for Analog Output[" + address + "]: " + result.result);
    })
}

exports.UpdateSnmpManager = function (snmp) {
    SnmpManager = snmp;
}

exports.UpdateSummitInfo = function (info) {
    SummitInfo = info;
}
//#endregion Analog Outputs Eventing
